#!/bin/bash
#
# Description:
#   Builds the PyCI package: minifies code, compresses pyci_static into a tarball, compresses plugins,
#   and then copies all the necessary files to the build directory.
#
# Author:
#   Dan McDougall <YouKnowWho@YouKnowWhat.com>
#
# TODO:  Make this also build an .egg and an .opkg.
# TODO:  Make this build PyDoc and JSDoc documentation.
# TODO:  Re-write this whole thing in Python.  It is getting unwieldly.
# TODO:  Arg, use a Makefile!


# NOTE:  Assumes we're in the same directory as PyCI!

PYCI_DIR=`pwd`
BUILD_DIR="${PYCI_DIR}/build"
TEMP_DIR="${PYCI_DIR}/build_temp"
# Note:  The Python Minifier is not currently used due to too many bugs
PYMINIFY="python /home/riskable/workspace/embedded_tools/pyminifier.py"
YUICOMPRESSOR="java -jar ${PYCI_DIR}/yuicompressor-2.4.2/build/yuicompressor-2.4.2.jar"
PYBABEL="/usr/bin/pybabel"
# This is used to build the ipk (opkg)
OPENWRT_TRUNK="/storage/openwrt-routerstation/openwrt"

# Remove the build dir and the temporary dir then re-make them
rm -rf ${BUILD_DIR}
rm -rf ${TEMP_DIR}
mkdir -p ${TEMP_DIR}
mkdir -p ${BUILD_DIR}/plugins_enabled
mkdir -p ${BUILD_DIR}/plugins_disabled

# Clean up .pyc files (so we don't accidentally copy or compress them)
find . -name '*.pyc' -delete > /dev/null

# Clean up any .directory files
find . -name '.directory' -delete > /dev/null

# Remove the minified/combined versions of our static js and css files
rm -f static/js/combined.js
rm -f static/css/combined.css

# Extract translatable texts from PyCI and plugins then update and compilethem.  Note: Assumes all locales have already been initialized
${PYBABEL} extract -o i18n/pyci.pot -F babel_pyci.cfg --no-location --msgid-bugs-address=riskable@youknowwhat.com . > /dev/null 2>&1
for locale in `ls -l i18n | egrep '^d' | awk '{print $8}'` ; do
    ${PYBABEL} update -D PyCI -i i18n/pyci.pot -d i18n/ -l $locale > /dev/null 2>&1
done
${PYBABEL} compile -D PyCI -d i18n/ -f --statistics > /dev/null 2>&1
cp -R i18n ${BUILD_DIR}/
rm -f ${BUILD_DIR}/i18n/*.pot # Only the .mo files are used by the running application
rm -f ${BUILD_DIR}/i18n/*/*/*.po # FYI: .mo files are compiled from .po files
# Make a note of which languages we have so we can clean up excess jqGrid locale files at the end
LANGUAGES=`ls ${BUILD_DIR}/i18n/`
LANGUAGES="en_US ${LANGUAGES}"

for plugin in `ls plugins_enabled` ; do
    cd ${PYCI_DIR}/plugins_enabled/${plugin}
    if [[ -d "i18n" ]]; then
        ${PYBABEL} extract -o i18n/${plugin}.pot -F ${PYCI_DIR}/babel_plugins.cfg --no-location --msgid-bugs-address=riskable@youknowwhat.com . > /dev/null 2>&1
        for locale in `ls -l i18n | egrep '^d' | awk '{print $8}'` ; do
            ${PYBABEL} update -D ${plugin} -i i18n/${plugin}.pot -d i18n/ -l $locale > /dev/null 2>&1
        done
        ${PYBABEL} compile -D ${plugin} -d i18n/ -f --statistics  > /dev/null 2>&1
    fi
done

cd ${PYCI_DIR}

# Record how much space we used before minifying and compressing
SPACE_BEFORE=`du -hs --exclude='build' --exclude='static.tgz'`

# Minifiy and combine the static javascript files
for filename in `ls uncombined/js/*.js` ; do
     echo "YUI Compressing: ${filename}"
    ${YUICOMPRESSOR} --nomunge ${filename} >> static/js/combined.js
# Debugging...
#     cat ${filename} >> static/js/combined.js
done
# Minifiy and combine the static css files
for filename in `ls uncombined/css/*.css` ; do
    echo "YUI Compressing: ${filename}"
    ${YUICOMPRESSOR} --nomunge ${filename} >> static/css/combined.css
done

# Copy the css images directory over as-is
cp -R uncombined/css/images static/css/ > /dev/null 2>&1

# Make a copy of the plugins_enabled and plugins_disabled directories under ${TEMP_DIR}
cp -R plugins_* ${TEMP_DIR}/ > /dev/null 2>&1

# Remove any copied svn directories and files
find ${TEMP_DIR} -name .svn -exec rm -rf {} \; > /dev/null 2>&1

# Minify all plugin css files (note: Not combining--that's up to the plugin authors)
for filepath in `find plugins_* -name *.css` ; do
    echo "YUI Compressing: ${filepath}"
    # This overwrites the copies of said files in ${TEMP_DIR}
    ${YUICOMPRESSOR} ${filepath} > ${TEMP_DIR}/${filepath}
done
# Minify all plugin js files
for filepath in `find plugins_* -name *.js` ; do
    echo "YUI Compressing: ${filepath}"
    # This overwrites the copies of said files in ${TEMP_DIR}
    ${YUICOMPRESSOR} --nomunge ${filepath} > ${TEMP_DIR}/${filepath}
done

# Minify and zip up plugins and place them in build/plugins (Note:  Just the .py files get zipped)
# ...and copy the i18n directories over minus the .pot files
# NOTE: This ends up in a wacky filestructure but it is only because i18n can't be inside a zip file
cd ${PYCI_DIR}/plugins_enabled/
for plugin in `ls`; do
    mkdir -p ${BUILD_DIR}/plugins_enabled/${plugin}
    for source_file in ${plugin}/*.py; do
#         ${PYMINIFY} ${source_file} > ${TEMP_DIR}/plugins_enabled/${source_file}
        cp ${source_file} ${TEMP_DIR}/plugins_enabled/${source_file}
    done
    cd ${TEMP_DIR}/plugins_enabled/${plugin}
    echo "Creating zip archive for: ${plugin}"
#     zip ${BUILD_DIR}/plugins_enabled/${plugin}.zip -r . -X -x \*.pyc -x \*.svn -x i18n > /dev/null
    zip ${BUILD_DIR}/plugins_enabled/${plugin}.zip -r -X *.py static templates -x \*.pyc -x \*.svn > /dev/null
#     zip ${BUILD_DIR}/plugins_enabled/${plugin}.zip *.py > /dev/null
#     tar -zcf ${BUILD_DIR}/plugins_enabled/${plugin}/static.tgz static
#     [[ -d "templates" ]] && cp -R templates ${BUILD_DIR}/plugins_enabled/${plugin}/
    [[ -d "i18n" ]] && cp -R i18n ${BUILD_DIR}/plugins_enabled/${plugin}/
    rm -f ${BUILD_DIR}/plugins_enabled/${plugin}/i18n/*.pot
    rm -f ${BUILD_DIR}/plugins_enabled/${plugin}/i18n/*/*/*.po
    cd ${PYCI_DIR}/plugins_enabled/
done
cd ${PYCI_DIR}/plugins_disabled/
for plugin in `ls`; do
    mkdir -p ${BUILD_DIR}/plugins_disabled/${plugin}
    for source_file in ${plugin}/*.py; do
#         ${PYMINIFY} ${source_file} > ${TEMP_DIR}/plugins_disabled/${source_file}
        cp ${source_file} ${TEMP_DIR}/plugins_disabled/${source_file}
    done
    cd ${TEMP_DIR}/plugins_disabled/${plugin}
    echo "Creating zip archive for: ${plugin}"
#     zip ${BUILD_DIR}/plugins_disabled/${plugin}.zip -r . -X -x \*.pyc -x \*.svn > /dev/null
    zip ${BUILD_DIR}/plugins_disabled/${plugin}.zip -r -X *.py static templates -x \*.pyc -x \*.svn > /dev/null
#     zip ${BUILD_DIR}/plugins_disabled/${plugin}.zip *.py > /dev/null
#     tar -zcf ${BUILD_DIR}/plugins_disabled/${plugin}/static.tgz static
#     [[ -d "templates" ]] && cp -R templates ${BUILD_DIR}/plugins_disabled/${plugin}/
    [[ -d "i18n" ]] && cp -R i18n ${BUILD_DIR}/plugins_disabled/${plugin}/
    rm -f ${BUILD_DIR}/plugins_disabled/${plugin}/i18n/*.pot
    rm -f ${BUILD_DIR}/plugins_disabled/${plugin}/i18n/*/*/*.po
    cd ${PYCI_DIR}/plugins_disabled/
done

cd ${PYCI_DIR}

# Make a copy of the themes directory in TEMP_DIR
cp -R themes ${TEMP_DIR}/ > /dev/null 2>&1

# Remove any copied svn directories and files
find ${TEMP_DIR} -name .svn -exec rm -rf {} \; > /dev/null 2>&1

# Minify all theme css files
for filepath in `find themes -name *.css` ; do
    echo "YUI Compressing: ${filepath}"
    # This overwrites the copies of said files in ${BUILD_DIR}
    ${YUICOMPRESSOR} ${filepath} > ${TEMP_DIR}/${filepath}
done
# Minify all theme js files
for filepath in `find themes -name *.js` ; do
    echo "YUI Compressing: ${filepath}"
    # This overwrites the copies of said files in ${BUILD_DIR}
    ${YUICOMPRESSOR} --nomunge ${filepath} > ${TEMP_DIR}/${filepath}
done

# NOTE: This doesn't work and CAN'T work because mako can't do template inheretance from within a tarball or zip file
# TODO:  Find a replacement for Mako that can do template inheretance from inside an archive (if not, write one!)
# Make the theme tarballs
# mkdir -p ${BUILD_DIR}/themes
# cd ${TEMP_DIR}/themes
# for theme in `ls`; do
#     cd ${TEMP_DIR}/themes/${theme}
#     tar -zcf ${BUILD_DIR}/themes/${theme}.tgz * > /dev/null
# done

mkdir -p ${BUILD_DIR}/themes
cd ${TEMP_DIR}/themes
for theme in `ls`; do
    echo "Compressing (gzip) static files for theme: ${theme}"
#     mkdir -p ${BUILD_DIR}/themes/${theme}/static
    cd ${TEMP_DIR}/themes/${theme}/static
    # Changing this to use gzip-compressed files instead of storing them inside of a tarball
#     tar -zcf ${BUILD_DIR}/themes/${theme}/static.tgz static > /dev/null
    gzip -r *
    cp -R ${TEMP_DIR}/themes/${theme} ${BUILD_DIR}/themes/
    # Now get rid of the .gz extension
#     for filename in `find . -name \*.gz`; do
#         NEWFILENAME=`echo $filename | sed -e 's/.gz$//g'`
#         mv $filename $NEWFILENAME
#     done
done

cd ${PYCI_DIR}

# Remove the leftover directories leaving only the zip files in the plugin folders
# find ${BUILD_DIR}/plugins_enabled/ -depth -mindepth 1 -type d -exec rm -rf {} \; > /dev/null

# Shrinksafe the combined js
# java -jar shrinksafe.jar /tmp/combined.js > static/js/combined.js

# Copy the cherrypy.conf over
cp cherrypy.conf ${BUILD_DIR}/

cp -R static ${TEMP_DIR}/ > /dev/null 2>&1
find ${TEMP_DIR} -name .svn -exec rm -rf {} \; > /dev/null 2>&1
cd ${TEMP_DIR}/static
# Disabling this since I'm now using individually-gzipped files
# tar -zcf ${BUILD_DIR}/static.tgz * > /dev/null
# Gzip all files in static individually then copy them to the build directory...
gzip -r *
mkdir -p ${BUILD_DIR}/static
cp -R * ${BUILD_DIR}/static/

# Now clean up any unneeded jqGrid locale files...
cd ${BUILD_DIR}/static/js/i18n
mkdir -p ${TEMP_DIR}/jqgrid
for language in $LANGUAGES; do
    # We have to convert our long-named $LANGUAGES to two-letter names
    language=`echo ${language} | cut -c1-2`
    for filename in `ls`; do
        if [[ "${filename}" = "grid.locale-${language}.js.gz" ]]; then
            # Move to our jqgrid temporary dir
            mv ${filename} ${TEMP_DIR}/jqgrid/
        fi
    done
done
# Remove all locale files not preserved in the temp directories
rm ${BUILD_DIR}/static/js/i18n/*
# Move the saved locales back
mv ${TEMP_DIR}/jqgrid/* ${BUILD_DIR}/static/js/i18n/

cd ${PYCI_DIR}

# Make a copy for local testing
# cp ${BUILD_DIR}/static.tgz ${PYCI_DIR}

#
# NOTE: Python minification disabled for now
#
# TODO:  Make the minifier work.  It is presently only stripping comments
# Minifiy and copy the .py files
# ${PYMINIFY} PyCI.py > ${BUILD_DIR}/PyCI.py
# mkdir -p ${TEMP_DIR}/sources
# for source_file in *.py; do
#     ${PYMINIFY} ${source_file} > ${TEMP_DIR}/sources/${source_file}
# done
cp PyCI.py ${BUILD_DIR}/PyCI.py
mkdir -p ${TEMP_DIR}/sources
for source_file in *.py; do
    cp ${source_file} ${TEMP_DIR}/sources/${source_file}
done
# Now zip up all the sources
cd ${TEMP_DIR}/sources
zip ${BUILD_DIR}/pyci_modules.zip *.py -x PyCI.py > /dev/null

# Copy over the required_modules directory
cp -R ${PYCI_DIR}/required_modules ${BUILD_DIR}/

# Copy over the metadata directory
cp -R ${PYCI_DIR}/metadata ${BUILD_DIR}/

# cd ${PYCI_DIR}/plugins_enabled
# for plugin in *; do
#     zip -r ${plugin}.zip ${plugin}
# done
# mv *.zip ${BUILD_DIR}/plugins_enabled/

# Copy the rest (assumes themes don't need minification or compression)
cd ${PYCI_DIR}
# cp -R themes ${BUILD_DIR} > /dev/null
cp key.pem certificate.pem ${BUILD_DIR}
# Don't need the static dir since we already compressed it
# rm -rf ${BUILD_DIR}/themes/*/static

# Make the symbolic link for html if it doesn't already exist
[[ ! -f ${BUILD_DIR}/html ]] && cd ${BUILD_DIR}; ln -s themes/routerstation html

cd ${PYCI_DIR}
chmod 755 ${BUILD_DIR}/PyCI.py

# Remove any copied svn directories and files
find ${BUILD_DIR} -name .svn -exec rm -rf {} \; > /dev/null

# Record how much space is used after minifying and compressing
SPACE_AFTER=`du -hs build`

# Clean up the TEMP_DIR
rm -rf ${TEMP_DIR}

# Copy the build dir to the OpenWRT package directory (must be copied--can't be linked)
cp -R ${BUILD_DIR}/* ${PYCI_DIR}/opkg/files/usr/PyCI/

# Now build the ipk if $OPENWRT_TRUNK exists
if [[ -d ${OPENWRT_TRUNK} ]]; then
    echo "Building ipk using: ${OPENWRT_TRUNK}"
    # Make sure the package symbolic link is present
    if [[ -L ${OPENWRT_TRUNK}/package/PyCI ]]; then
        # Build the ipk
        cd ${OPENWRT_TRUNK}
        make package/PyCI/{clean,compile,install}
    else
        # Make the link
        ln -s ${PYCI_DIR}/opkg ${OPENWRT_TRUNK}/package/PyCI
        # Build the ipk
        cd ${OPENWRT_TRUNK}
        make package/PyCI/{clean,compile,install}
    fi
    IPK_SIZE=`ls -lh ${OPENWRT_TRUNK}/bin/packages/*/PyCI*.ipk | awk '{print $5}'`
    # Update the Packages and Packages.gz files with the proper md5sum
    cd ${OPENWRT_TRUNK}/bin/packages/*/
    ${OPENWRT_TRUNK}/scripts/ipkg-make-index.sh . > Packages
    gzip -9c Packages > Packages.gz
fi

echo "SPACE BEFORE: ${SPACE_BEFORE}"
echo "SPACE AFTER: ${SPACE_AFTER}"
echo "IPK SIZE: ${IPK_SIZE}"