# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

import subprocess, sys
from optparse import OptionParser
sys.path.append('/home/riskable/workspace/routerstation/PyCI')
from configfile import ConfigFile
from mako.template import Template

# This is a mako template for outputting iptables lines from our Rule() objects
rule_template = """
<%!
def s(text):
    return repr(unicode(text).encode('utf-8'))

def optional(switch, value):
    if value is not None:
        if value == True:
            return switch
        else:
            return '%s %s' % (switch, s(value))
    return ''
%>
% for rule in rules:
iptables -t filter \\
% if rule.ruleno:
-I ${rule.chain|s} ${rule.ruleno|s} \\
% else:
-A ${rule.chain|s} \\
% endif
% if isinstance(match, list):
 % for item in rule.match:
${optional('--match', item)} \\
 % endfor
% else:
${optional('--match', rule.match)} \\
% endif
${optional('--source', rule.source)} \\
${optional('--destination', rule.destination)} \\
${optional('--protocol', rule.protocol)} \\
${optional('--interface', rule.in_interface)} \\
${optional('--out-interface', rule.out_interface)} \\
${optional('--rcheck', rule.rcheck)} \\
${optional('--modprobe', rule.modprobe)} \\
${optional('--src-type', rule.src_type)} \\
${optional('--dst-type', rule.dst_type)} \\
${optional('--limit-iface_in', rule.limit_iface_in)} \\
${optional('--limit-iface_out', rule.limit_iface_out)} \\
${optional('--ahspi', rule.ahspi)} \\
${optional('--comment', rule.comment)} \\
${optional('--connbytes', rule.connbytes)} \\
${optional('--connbytes-dir', rule.connbytes_dir)} \\
${optional('--connlimit-above', rule.connlimit_above)} \\
${optional('--connlimit-mask', rule.connlimit_mask)} \\
${optional('--mark', rule.mark)} \\
${optional('--ctstate', rule.ctstate)} \\
${optional('--ctproto', rule.ctproto)} \\
${optional('--ctorigsrc', rule.ctorigsrc)} \\
${optional('--ctorigdst', rule.ctorigdst)} \\
${optional('--ctreplsrc', rule.ctreplsrc)} \\
${optional('--ctrepldst', rule.ctrepldst)} \\
${optional('--ctorigsrcport', rule.ctorigsrcport)} \\
${optional('--ctorigdstport', rule.ctorigdstport)} \\
${optional('--ctreplsrcport', rule.ctreplsrcport)} \\
${optional('--ctrepldstport', rule.ctrepldstport)} \\
${optional('--ctstatus', rule.ctstatus)} \\
${optional('--ctexpire', rule.ctexpire)} \\
${optional('--ctdir', rule.ctdir)} \\
${optional('--source-port', rule.source_port)} \\
${optional('--source-ports', rule.source_ports)} \\
${optional('--destination-port', rule.destination_port)} \\
${optional('--destination-ports', rule.destination_ports)} \\
${optional('--dccp-types', rule.dccp_types)} \\
${optional('--dccp-option', rule.dccp_option)} \\
${optional('--dscp', rule.dscp)} \\
${optional('--dscp-class', rule.dscp_class)} \\
${optional('--ecn-tcp-cwr', rule.ecn_tcp_cwr)} \\
${optional('--ecn-tcp-ece', rule.ecn_tcp_ece)} \\
${optional('--ecn-ip-ect', rule.ecn_ip_ect)} \\
${optional('--espspi', rule.espspi)} \\
${optional('--hashlimit-upto', rule.hashlimit_upto)} \\
${optional('--hashlimit-above', rule.hashlimit_above)} \\
${optional('--hashlimit-name', rule.hashlimit_name)} \\
${optional('--hashlimit-burst', rule.hashlimit_burst)} \\
${optional('--hashlimit-mode', rule.hashlimit_mode)} \\
${optional('--hashlimit-srcmask', rule.hashlimit_srcmask)} \\
${optional('--hashlimit-dstmask', rule.hashlimit_dstmask)} \\
${optional('--hashlimit-htable_size', rule.hashlimit_htable_size)} \\
${optional('--hashlimit-htable_max', rule.hashlimit_htable_max)} \\
${optional('--hashlimit-htable_expire', rule.hashlimit_htable_expire)} \\
${optional('--hashlimit-htable_gcinterval', rule.hashlimit_htable_gcinterval)} \\
${optional('--helper', rule.helper)} \\
${optional('--icmp-type', rule.icmp_type)} \\
${optional('--src-range', rule.src_range)} \\
${optional('--dst-range', rule.dst_range)} \\
${optional('--length', rule.length)} \\
${optional('--limit', rule.limit)} \\
${optional('--limit-burst', rule.limit_burst)} \\
${optional('--mac-source', rule.mac_source)} \\
${optional('--ports', rule.ports)} \\
${optional('--uid-owner', rule.uid_owner)} \\
${optional('--gid-owner', rule.gid_owner)} \\
${optional('--socket-exists', rule.socket_exists)} \\
${optional('--physdev-in', rule.physdev_in)} \\
${optional('--physdev-out', rule.physdev_out)} \\
${optional('--physdev-is-in', rule.physdev_is_in)} \\
${optional('--physdev-is-out', rule.physdev_is_out)} \\
${optional('--physdev-is-bridged', rule.physdev_is_bridged)} \\
${optional('--pkg-type', rule.pkg_type)} \\
${optional('--dir', rule.dir_)} \\
${optional('--pol', rule.pol)} \\
${optional('--strict', rule.strict)} \\
${optional('--reqid', rule.reqid)} \\
${optional('--spi', rule.spi)} \\
${optional('--proto', rule.proto)} \\
${optional('--mode', rule.mode)} \\
${optional('--tunnel-src', rule.tunnel_src)} \\
${optional('--tunnel-dst', rule.tunnel_dst)} \\
${optional('--next', rule.next)} \\
${optional('--quota', rule.quota)} \\
${optional('--realm', rule.realm)} \\
${optional('--name', rule.name)} \\
${optional('--set', rule.set_)} \\
${optional('--rcheck', rule.rcheck)} \\
${optional('--update', rule.update)} \\
${optional('--remove', rule.remove)} \\
${optional('--seconds', rule.seconds)} \\
${optional('--hitcount', rule.hitcount)} \\
${optional('--rttl', rule.rttl)} \\
${optional('--rsource', rule.rsource)} \\
${optional('--rdest', rule.rdest)} \\
${optional('--chunk-types', rule.chunk_types)} \\
${optional('--state', rule.state)} \\
${optional('--probability', rule.probability)} \\
${optional('--every', rule.every)} \\
${optional('--packet', rule.packet)} \\
${optional('--algo', rule.algo)} \\
${optional('--from', rule.from_)} \\
${optional('--to', rule.to)} \\
${optional('--string', rule.string)} \\
${optional('--hex-string', rule.hex_string)} \\
${optional('--tcp-flags', rule.tcp_flags)} \\
${optional('--syn', rule.syn)} \\
${optional('--tcp-option', rule.tcp_option)} \\
${optional('--mss', rule.mss)} \\
${optional('--datestart', rule.datestart)} \\
${optional('--datestop', rule.datestop)} \\
${optional('--timestart', rule.timestart)} \\
${optional('--timestop', rule.timestop)} \\
${optional('--monthdays', rule.monthdays)} \\
${optional('--weekdays', rule.weekdays)} \\
${optional('--utc', rule.utc)} \\
${optional('--localtz', rule.localtz)} \\
${optional('--tos', rule.tos)} \\
${optional('--ttl-eq', rule.ttl_eq)} \\
${optional('--ttl-gt', rule.ttl_gt)} \\
${optional('--ttl-lt', rule.ttl_lt)} \\
${optional('--u32', rule.u32)} \\
${optional('--set-class', rule.set_class)} \\
${optional('--new', rule.new)} \\
${optional('--hashmode', rule.hashmode)} \\
${optional('--clustermac', rule.clustermac)} \\
${optional('--total-nodes', rule.total_nodes)} \\
${optional('--local-nodes', rule.local_nodes)} \\
${optional('--hash-init', rule.hash_init)} \\
${optional('--set-xmark', rule.set_xmark)} \\
${optional('--save-mark', rule.save_mark)} \\
${optional('--restore-mark', rule.restore_mark)} \\
${optional('--and-mark', rule.and_mark)} \\
${optional('--or-mark', rule.or_mark)} \\
${optional('--xor-mark', rule.xor_mark)} \\
${optional('--to-destination', rule.to_destination)} \\
${optional('--random', rule.random)} \\
${optional('--set-dscp', rule.set_dscp)} \\
${optional('--set-dscp-class', rule.set_dscp_class)} \\
${optional('--ecn-tcp-remove', rule.ecn_tcp_remove)} \\
${optional('--log-level', rule.log_level)} \\
${optional('--log-prefix', rule.log_prefix)} \\
${optional('--log-tcp-sequence', rule.log_tcp_sequence)} \\
${optional('--log-tcp-options', rule.log_tcp_options)} \\
${optional('--log-ip-options', rule.log_ip_options)} \\
${optional('--log-uid', rule.log_uid)} \\
${optional('--to-ports', rule.to_ports)} \\
${optional('--nflog-group', rule.nflog_group)} \\
${optional('--nflog-prefix', rule.nflog_prefix)} \\
${optional('--nflog-range', rule.nflog_range)} \\
${optional('--nflog-threshold', rule.nflog_threshold)} \\
${optional('--queue-num', rule.queue_num)} \\
${optional('--ratetest-name', rule.ratetest_name)} \\
${optional('--ratetest-interval', rule.ratetest_interval)} \\
${optional('--ratetest-ewmalog', rule.ratetest_ewmalog)} \\
${optional('--nodst', rule.nodst)} \\
${optional('--selectx', rule.selectx)} \\
${optional('--add-set', rule.add_set)} \\
${optional('--del-set', rule.del_set)} \\
${optional('--to-source', rule.to_source)} \\
${optional('--set-mss', rule.set_mss)} \\
${optional('--clamp-mss-to-pmtu', rule.clamp_mss_to_pmtu)} \\
${optional('--strip-options', rule.strip_options)} \\
${optional('--set-tos', rule.set_tos)} \\
${optional('--and-tos', rule.and_tos)} \\
${optional('--or-tos', rule.or_tos)} \\
${optional('--xor-tos', rule.xor_tos)} \\
${optional('--ttl-set', rule.ttl_set)} \\
${optional('--ttl-dec', rule.ttl_dec)} \\
${optional('--ttl-inc', rule.ttl_inc)} \\
${optional('--ulog-nlgroup', rule.ulog_nlgroup)} \\
${optional('--ulog-prefix', rule.ulog_prefix)} \\
${optional('--ulog-cprange', rule.ulog_cprange)} \\
${optional('--ulog-qthreshold', rule.ulog_qthreshold)} \\
${optional('--ipp2p', rule.ipp2p)} \\
${optional('--edk', rule.edonkey)} \\
${optional('--kazaa', rule.kazaa)} \\
${optional('--gnu', rule.gnutella)} \\
${optional('--dc', rule.directconnect)} \\
${optional('--bit', rule.bittorrent)} \\
${optional('--apple', rule.apple)} \\
${optional('--winmx', rule.winmx)} \\
${optional('--soul', rule.soulseek)} \\
${optional('--ares', rule.ares)} \\
-j ${rule.target|s} \\
${optional('--reject-with', rule.reject_with)}
## TODO:  Finish writing all these switches (and figure out if they need to go in a certain order)
%endfor"""

def parse_iptables_args(args):
    """Parses iptables arguments, 'args' (which should be the equivalent to sys.argv[1:]).
    Returns options (as output by OptionParser)."""
    parser = OptionParser()
    parser.disable_interspersed_args()
    # Man, iptables has a lot of arguments...
    parser.add_option("-t", "--table", dest="table", default='filter')
    parser.add_option("-A", "--append", dest="chain", default=None)
    parser.add_option("-D", "--delete", dest="delete", default=None)
    parser.add_option("-I", "--insert", dest="insert", default=None)
    parser.add_option("-R", "--replace", dest="replace", default=None)
    parser.add_option("-L", "--list", dest="list", default=None)
    #parser.add_option("-S", "--list-rules", dest="list_rules", default=True) # Not sure if I need this--more of an interactive thing
    parser.add_option("-F", "--flush", dest="flush", default=None)
    parser.add_option("-Z", "--zero", dest="zero", default=None)
    parser.add_option("-N", "--new-chain", dest="new_chain", default=None)
    parser.add_option("-X", "--delete-chain", dest="delete_chain", default=None)
    parser.add_option("-P", "--policy", dest="policy", default=None)
    parser.add_option("-E", "--rename_chain", dest="rename_chain", default=None)
    parser.add_option("-p", "--protocol", dest="protocol", default="all")
    parser.add_option("-s", "--source", dest="source", default="0.0.0.0/0")
    parser.add_option("-d", "--destination", dest="destination", default="0.0.0.0/0")
    parser.add_option("-j", "--jump", dest="target", default=None) # Note: "jump" is "target" in iptables -L so that's how I'm referring to it
    parser.add_option("-g", "--goto", dest="goto", default=None)
    parser.add_option("-i", "--interface", dest="in_interface", default=None)
    parser.add_option("-o", "--out-interface", dest="out_interface", default=None)
    parser.add_option("-f", "--fragment", dest="fragment", action="store_true", default=False)
    parser.add_option("-c", "--set-counters", dest="set_counters", default=None)
    parser.add_option("-m", "--match", dest="match", default=None)
    parser.add_option("--modprobe", dest="modprobe", default=None)
    parser.add_option("--src-type", dest="src_type", default=None)
    parser.add_option("--dst-type", dest="dst_type", default=None)
    parser.add_option("--limit-iface-in", dest="limit_iface_in", action="store_true", default=False)
    parser.add_option("--limit-iface-out", dest="limit_iface_out", action="store_true", default=False)
    parser.add_option("--ahspi", dest="ahspi", default=None)
    parser.add_option("--comment", dest="comment", default=None)
    parser.add_option("--connbytes", dest="connbytes", default=None)
    parser.add_option("--connbytes-dir", dest="connbytes_dir", default=None)
    parser.add_option("--connlimit-above", dest="connlimit_above", default=None)
    parser.add_option("--connlimit-mask", dest="connlimit_mask", default=None)
    parser.add_option("--mark", dest="mark", default=None)
    parser.add_option("--ctstate", dest="ctstate", default=None)
    parser.add_option("--ctproto", dest="ctproto", default=None)
    parser.add_option("--ctorigsrc", dest="ctorigsrc", default=None)
    parser.add_option("--ctorigdst", dest="ctorigdst", default=None)
    parser.add_option("--ctreplsrc", dest="ctreplsrc", default=None)
    parser.add_option("--ctrepldst", dest="ctrepldst", default=None)
    parser.add_option("--ctorigsrcport", dest="ctorigsrcport", default=None)
    parser.add_option("--ctorigdstport", dest="ctorigdstport", default=None)
    parser.add_option("--ctreplsrcport", dest="ctreplsrcport", default=None)
    parser.add_option("--ctrepldstport", dest="ctrepldstport", default=None)
    parser.add_option("--ctstatus", dest="ctstatus", default=None)
    parser.add_option("--ctexpire", dest="ctexpire", default=None)
    parser.add_option("--ctdir", dest="ctdir", default=None)
    parser.add_option("--source-port", "--sport", dest="source_port", default=None)
    parser.add_option("--source-ports", "--sports", dest="source_ports", default=None)
    parser.add_option("--destination-port", "--dport", dest="destination_port", default=None)
    parser.add_option("--destination-ports", "--dports", dest="destination_ports", default=None)
    parser.add_option("--dccp-types", dest="dccp_types", default=None)
    parser.add_option("--dccp-option", dest="dccp_option", default=None)
    parser.add_option("--dscp", dest="dscp", default=None)
    parser.add_option("--dscp-class", dest="dscp_class", default=None)
    parser.add_option("--ecn-tcp-cwr", dest="ecn_tcp_cwr", action="store_true", default=False)
    parser.add_option("--ecn-tcp-ece", dest="ecn_tcp_ece", action="store_true", default=False)
    parser.add_option("--ecn-ip-ect", dest="ecn_ip_ect", default=None)
    parser.add_option("--espspi", dest="espspi", default=None)
    parser.add_option("--hashlimit-upto", dest="hashlimit_upto", default=None)
    parser.add_option("--hashlimit-above", dest="hashlimit_above", default=None)
    parser.add_option("--hashlimit-name", dest="hashlimit_name", default=None)
    parser.add_option("--hashlimit-burst", dest="hashlimit_burst", default=None)
    parser.add_option("--hashlimit-mode", dest="hashlimit_mode", default=None)
    parser.add_option("--hashlimit-srcmask", dest="hashlimit_srcmask", default=None)
    parser.add_option("--hashlimit-dstmask", dest="hashlimit_dstmask", default=None)
    parser.add_option("--hashlimit-htable-size", dest="hashlimit_htable_size", default=None)
    parser.add_option("--hashlimit-htable-max", dest="hashlimit_htable_max", default=None)
    parser.add_option("--hashlimit-htable-expire", dest="hashlimit_htable_expire", default=None)
    parser.add_option("--hashlimit-htable-gcinterval", dest="hashlimit_htable_gcinterval", default=None)
    parser.add_option("--helper", dest="helper", default=None)
    parser.add_option("--icmp-type", dest="icmp_type", default=None)
    parser.add_option("--src-range", dest="src_range", default=None)
    parser.add_option("--dst-range", dest="dst_range", default=None)
    parser.add_option("--length", dest="length", default=None)
    parser.add_option("--limit", dest="limit", default=None)
    parser.add_option("--limit-burst", dest="limit_burst", default=None)
    parser.add_option("--mac-source", dest="mac_source", default=None)
    parser.add_option("--ports", dest="ports", default=None)
    parser.add_option("--uid-owner", dest="uid_owner", default=None)
    parser.add_option("--gid-owner", dest="gid_owner", default=None)
    parser.add_option("--socket-exists", dest="socket_exists", action="store_true", default=False)
    parser.add_option("--physdev-in", dest="physdev_in", default=None)
    parser.add_option("--physdev-out", dest="physdev_out", default=None)
    parser.add_option("--physdev-is-in", dest="physdev_is_in", action="store_true", default=False)
    parser.add_option("--physdev-is-out", dest="physdev_is_out", action="store_true", default=False)
    parser.add_option("--physdev-is-bridged", dest="physdev_is_bridged", action="store_true", default=False)
    parser.add_option("--pkt-type", dest="pkg_type", default=None)
    parser.add_option("--dir", dest="dir_", default=None) # dir is a reserved word
    parser.add_option("--pol", dest="pol", default=None)
    parser.add_option("--strict", dest="strict", action="store_true", default=False)
    parser.add_option("--reqid", dest="reqid", default=None)
    parser.add_option("--spi", dest="spi", default=None)
    parser.add_option("--proto", dest="proto", default=None)
    parser.add_option("--mode", dest="mode", default=None)
    parser.add_option("--tunnel-src", dest="tunnel_src", default=None)
    parser.add_option("--tunnel-dst", dest="tunnel_dst", default=None)
    parser.add_option("--next", dest="next", action="store_true", default=False)
    parser.add_option("--quota", dest="quota", default=None)
    parser.add_option("--realm", dest="realm", default=None)
    # That's 100 arguments...
    parser.add_option("--name", dest="name", default=None)
    parser.add_option("--set", dest="set_", action="store_true", default=False) # set is a reserved word
    parser.add_option("--rcheck", dest="rcheck", action="store_true", default=False)
    parser.add_option("--update", dest="update", action="store_true", default=False)
    parser.add_option("--remove", dest="remove", action="store_true", default=False)
    parser.add_option("--seconds", dest="seconds", default=None)
    parser.add_option("--hitcount", dest="hitcount", default=None)
    parser.add_option("--rttl", dest="rttl", action="store_true", default=False)
    parser.add_option("--rsource", dest="rsource", action="store_true", default=False)
    parser.add_option("--rdest", dest="rdest", action="store_true", default=False)
    parser.add_option("--chunk-types", dest="chunk_types", default=None)
    parser.add_option("--state", dest="state", default=None)
    parser.add_option("--probability", dest="probability", default=None)
    parser.add_option("--every", dest="every", default=None)
    parser.add_option("--packet", dest="packet", default=None)
    parser.add_option("--algo", dest="algo", default=None)
    parser.add_option("--from", dest="from_", default=None) # From is reserved too
    parser.add_option("--to", dest="to", default=None)
    parser.add_option("--string", dest="string", default=None)
    parser.add_option("--hex-string", dest="hex_string", default=None)
    parser.add_option("--tcp-flags", dest="tcp_flags", default=None)
    parser.add_option("--syn", dest="syn", action="store_true", default=False)
    parser.add_option("--tcp-option", dest="tcp_option", default=None)
    parser.add_option("--mss", dest="mss", default=None)
    parser.add_option("--datestart", dest="datestart", default=None)
    parser.add_option("--datestop", dest="datestop", default=None)
    parser.add_option("--timestart", dest="timestart", default=None)
    parser.add_option("--timestop", dest="timestop", default=None)
    parser.add_option("--monthdays", dest="monthdays", default=None)
    parser.add_option("--weekdays", dest="weekdays", default=None)
    parser.add_option("--utc", dest="utc", action="store_true", default=False)
    parser.add_option("--localtz", dest="localtz", action="store_true", default=False)
    parser.add_option("--tos", dest="tos", default=None)
    parser.add_option("--ttl-eq", dest="ttl_eq", default=None)
    parser.add_option("--ttl-gt", dest="ttl_gt", default=None)
    parser.add_option("--ttl-lt", dest="ttl_lt", default=None)
    parser.add_option("--u32", dest="u32", default=None)
    parser.add_option("--set-class", dest="set_class", default=None)
    parser.add_option("--new", dest="new", action="store_true", default=False)
    parser.add_option("--hashmode", dest="hashmode", default=None)
    parser.add_option("--clustermac", dest="clustermac", default=None)
    parser.add_option("--total-nodes", dest="total_nodes", default=None)
    parser.add_option("--local_nodes", dest="local_nodes", default=None)
    parser.add_option("--hash-init", dest="hash_init", default=None)
    parser.add_option("--set-xmark", dest="set_xmark", default=None)
    parser.add_option("--save-mark", dest="save_mark", default=None)
    parser.add_option("--restore-mark", dest="restore_mark", default=None)
    parser.add_option("--and-mark", dest="and_mark", default=None)
    parser.add_option("--or-mark", dest="or_mark", default=None)
    parser.add_option("--xor-mark", dest="xor_mark", default=None)
    #parser.add_option("--save", dest="save", action="store_true", default=False) # Don't need these
    #parser.add_option("--restore", dest="restore", action="store_true", default=False)
    parser.add_option("--to-destination", dest="to_destination", default=None)
    parser.add_option("--random", dest="random", action="store_true", default=False)
    parser.add_option("--set-dscp", dest="set_dscp", default=None)
    parser.add_option("--set-dscp-class", dest="set_dscp_class", default=None)
    parser.add_option("--ecn-tcp-remove", dest="ecn_tcp_remove", action="store_true", default=False)
    parser.add_option("--log-level", dest="log_level", default=None)
    parser.add_option("--log-prefix", dest="log_prefix", default=None)
    parser.add_option("--log-tcp-sequence", dest="log_tcp_sequence", action="store_true", default=False)
    parser.add_option("--log-tcp-options", dest="log_tcp_options", action="store_true", default=False)
    parser.add_option("--log-ip-options", dest="log_ip_options", action="store_true", default=False)
    parser.add_option("--log-uid", dest="log_uid", action="store_true", default=False)
    parser.add_option("--to-ports", dest="to_ports", default=None)
    parser.add_option("--nflog-group", dest="nflog_group", default=None)
    parser.add_option("--nflog-prefix", dest="nflog_prefix", default=None)
    parser.add_option("--nflog-range", dest="nflog_range", default=None)
    parser.add_option("--nflog-threshold", dest="nflog_threshold", default=None)
    parser.add_option("--queue-num", dest="queue_num", default=None)
    parser.add_option("--ratetest-name", dest="ratetest_name", default=None)
    parser.add_option("--ratetest-interval", dest="ratetest_interval", default=None)
    parser.add_option("--ratetest-ewmalog", dest="ratetest_ewmalog", default=None)
    parser.add_option("--reject-with", dest="reject_with", default=None)
    parser.add_option("--nodst", dest="nodst", action="store_true", default=False)
    parser.add_option("--selectx", dest="selectx", default=None)
    parser.add_option("--add-set", dest="add_set", default=None)
    parser.add_option("--del-set", dest="del_set", default=None)
    parser.add_option("--to-source", dest="to_source", default=None)
    parser.add_option("--set-mss", dest="set_mss", default=None)
    parser.add_option("--clamp-mss-to-pmtu", dest="clamp_mss_to_pmtu", default=None)
    parser.add_option("--strip-options", dest="strip_options", default=None)
    parser.add_option("--set-tos", dest="set_tos", default=None)
    parser.add_option("--and-tos", dest="and_tos", default=None)
    parser.add_option("--or-tos", dest="or_tos", default=None)
    parser.add_option("--xor-tos", dest="xor_tos", default=None)
    parser.add_option("--ttl-set", dest="ttl_set", default=None)
    parser.add_option("--ttl-dec", dest="ttl_dec", default=None)
    parser.add_option("--ttl-inc", dest="ttl_inc", default=None)
    parser.add_option("--ulog-nlgroup", dest="ulog_nlgroup", default=None)
    parser.add_option("--ulog-prefix", dest="ulog_prefix", default=None)
    parser.add_option("--ulog-cprange", dest="ulog_cprange", default=None)
    parser.add_option("--ulog-qthreshold", dest="ulog_qthreshold", default=None)
    # 192 arguments in the standard iptables command!
    # IPP2P and related options...
    parser.add_option("--ipp2p", dest="ipp2p", action="store_true", default=False)
    parser.add_option("--edk", dest="edonkey", default=None)
    parser.add_option("--kazaa", dest="kazaa", default=None)
    parser.add_option("--gnu", dest="gnutella", default=None)
    parser.add_option("--dc", dest="directconnect", default=None)
    parser.add_option("--bit", dest="bittorrent", default=None)
    parser.add_option("--apple", dest="apple", default=None)
    parser.add_option("--winmx", dest="winmx", default=None)
    parser.add_option("--soul", dest="soulseek", default=None)
    parser.add_option("--ares", dest="ares", default=None)
    (options, args) = parser.parse_args(args=args)
    return options

class RuleParsingException(Exception): pass

class Rule(object):
    """Class to store IPTables rules"""
    def __init__(self,
        table,
        chain,
        status_params=None,
        match=None,
        target=None,
        protocol=None,
        target_opts=None,
        source=None,
        source_port=None,
        destination=None,
        destination_port=None,
        in_interface=None,
        out_interface=None,
        packets=None,
        bytes=None,
        ruleno=None,
        state=None,
        modprobe=None,
        src_type=None,
        dst_type=None,
        limit_iface_in=None,
        limit_iface_out=None,
        ahspi=None,
        comment=None,
        connbytes=None,
        connbytes_dir=None,
        connlimit_above=None,
        connlimit_mask=None,
        mark=None,
        ctstate=None,
        ctproto=None,
        ctorigsrc=None,
        ctorigdst=None,
        ctreplsrc=None,
        ctrepldst=None,
        ctorigsrcport=None,
        ctorigdstport=None,
        ctreplsrcport=None,
        ctrepldstport=None,
        ctstatus=None,
        ctexpire=None,
        ctdir=None,
        source_ports=None,
        destination_ports=None,
        dccp_types=None,
        dccp_option=None,
        dscp=None,
        dscp_class=None,
        ecn_tcp_cwr=None,
        ecn_tcp_ece=None,
        ecn_ip_ect=None,
        espspi=None,
        hashlimit_upto=None,
        hashlimit_above=None,
        hashlimit_name=None,
        hashlimit_burst=None,
        hashlimit_mode=None,
        hashlimit_srcmask=None,
        hashlimit_dstmask=None,
        hashlimit_htable_size=None,
        hashlimit_htable_max=None,
        hashlimit_htable_expire=None,
        hashlimit_htable_gcinterval=None,
        helper=None,
        icmp_type=None,
        src_range=None,
        dst_range=None,
        length=None,
        limit=None,
        limit_burst=None,
        mac_source=None,
        ports=None,
        uid_owner=None,
        gid_owner=None,
        socket_exists=None,
        physdev_in=None,
        physdev_out=None,
        physdev_is_in=None,
        physdev_is_out=None,
        physdev_is_bridged=None,
        pkg_type=None,
        dir_=None,
        pol=None,
        strict=None,
        reqid=None,
        spi=None,
        proto=None,
        mode=None,
        tunnel_src=None,
        tunnel_dst=None,
        next=None,
        quota=None,
        realm=None,
        name=None,
        set_=None,
        rcheck=None,
        update=None,
        remove=None,
        seconds=None,
        hitcount=None,
        rttl=None,
        rsource=None,
        rdest=None,
        chunk_types=None,
        probability=None,
        every=None,
        packet=None,
        algo=None,
        from_=None,
        to=None,
        string=None,
        hex_string=None,
        tcp_flags=None,
        syn=None,
        tcp_option=None,
        mss=None,
        datestart=None,
        datestop=None,
        timestart=None,
        timestop=None,
        monthdays=None,
        weekdays=None,
        utc=None,
        localtz=None,
        tos=None,
        ttl_eq=None,
        ttl_gt=None,
        ttl_lt=None,
        u32=None,
        set_class=None,
        new=None,
        hashmode=None,
        clustermac=None,
        total_nodes=None,
        local_nodes=None,
        hash_init=None,
        set_xmark=None,
        save_mark=None,
        restore_mark=None,
        and_mark=None,
        or_mark=None,
        xor_mark=None,
        to_destination=None,
        random=None,
        set_dscp=None,
        set_dscp_class=None,
        ecn_tcp_remove=None,
        log_level=None,
        log_prefix=None,
        log_tcp_sequence=None,
        log_tcp_options=None,
        log_ip_options=None,
        log_uid=None,
        to_ports=None,
        nflog_group=None,
        nflog_prefix=None,
        nflog_range=None,
        nflog_threshold=None,
        queue_num=None,
        ratetest_name=None,
        ratetest_interval=None,
        ratetest_ewmalog=None,
        reject_with=None,
        nodst=None,
        selectx=None,
        add_set=None,
        del_set=None,
        to_source=None,
        set_mss=None,
        clamp_mss_to_pmtu=None,
        strip_options=None,
        set_tos=None,
        and_tos=None,
        or_tos=None,
        xor_tos=None,
        ttl_set=None,
        ttl_dec=None,
        ttl_inc=None,
        ulog_nlgroup=None,
        ulog_prefix=None,
        ulog_cprange=None,
        ulog_qthreshold=None,
        ipp2p=None,
        edonkey=None,
        kazaa=None,
        gnutella=None,
        bittorrent=None,
        apple=None,
        winmx=None,
        directconnect=None,
        soulseek=None,
        ares=None
    ):
        self.table = table
        self.chain = chain
        self.match = match
        self.target = target
        self.target_opts = target_opts
        self.protocol = protocol
        self.source = source
        self.source_port = source_port
        self.destination = destination
        self.destination_port = destination_port
        self.in_interface = in_interface
        self.out_interface = out_interface
        self.packets = packets
        self.bytes = bytes
        self.ruleno = ruleno
        self.state = state
        self.ctstate = ctstate
        self.modprobe = modprobe
        self.src_type = src_type
        self.dst_type = dst_type
        self.limit_iface_in = limit_iface_in
        self.limit_iface_out = limit_iface_out
        self.ahspi = ahspi
        self.comment = comment
        self.connbytes = connbytes
        self.connbytes_dir = connbytes_dir
        self.connlimit_above = connlimit_above
        self.connlimit_mask = connlimit_mask
        self.mark = mark
        self.ctstate = ctstate
        self.ctproto = ctproto
        self.ctorigsrc = ctorigsrc
        self.ctorigdst = ctorigdst
        self.ctreplsrc = ctreplsrc
        self.ctrepldst = ctrepldst
        self.ctorigsrcport = ctorigsrcport
        self.ctorigdstport = ctorigdstport
        self.ctreplsrcport = ctreplsrcport
        self.ctrepldstport = ctrepldstport
        self.ctstatus = ctstatus
        self.ctexpire = ctexpire
        self.ctdir = ctdir
        self.source_ports = source_ports
        self.destination_ports = destination_ports
        self.dccp_types = dccp_types
        self.dccp_option = dccp_option
        self.dscp = dscp
        self.dscp_class = dscp_class
        self.ecn_tcp_cwr = ecn_tcp_cwr
        self.ecn_tcp_ece = ecn_tcp_ece
        self.ecn_ip_ect = ecn_ip_ect
        self.espspi = espspi
        self.hashlimit_upto = hashlimit_upto
        self.hashlimit_above = hashlimit_above
        self.hashlimit_name = hashlimit_name
        self.hashlimit_burst = hashlimit_burst
        self.hashlimit_mode = hashlimit_mode
        self.hashlimit_srcmask = hashlimit_srcmask
        self.hashlimit_dstmask = hashlimit_dstmask
        self.hashlimit_htable_size = hashlimit_htable_size
        self.hashlimit_htable_max = hashlimit_htable_max
        self.hashlimit_htable_expire = hashlimit_htable_expire
        self.hashlimit_htable_gcinterval = hashlimit_htable_gcinterval
        self.helper = helper
        self.icmp_type = icmp_type
        self.src_range = src_range
        self.dst_range = dst_range
        self.length = length
        self.limit = limit
        self.limit_burst = limit_burst
        self.mac_source = mac_source
        self.ports = ports
        self.uid_owner = uid_owner
        self.gid_owner = gid_owner
        self.socket_exists = socket_exists
        self.physdev_in = physdev_in
        self.physdev_out = physdev_out
        self.physdev_is_in = physdev_is_in
        self.physdev_is_out = physdev_is_out
        self.physdev_is_bridged = physdev_is_bridged
        self.pkg_type = pkg_type
        self.dir_ = dir_
        self.pol = pol
        self.strict = strict
        self.reqid = reqid
        self.spi = spi
        self.proto = proto
        self.mode = mode
        self.tunnel_src = tunnel_src
        self.tunnel_dst = tunnel_dst
        self.next = next
        self.quota = quota
        self.realm = realm
        self.name = name
        self.set_ = set_
        self.rcheck = rcheck
        self.update = update
        self.remove = remove
        self.seconds = seconds
        self.hitcount = hitcount
        self.rttl = rttl
        self.rsource = rsource
        self.rdest = rdest
        self.chunk_types = chunk_types
        self.probability = probability
        self.every = every
        self.packet = packet
        self.algo = algo
        self.from_ = from_
        self.to = to
        self.string = string
        self.hex_string = hex_string
        self.tcp_flags = tcp_flags
        self.syn = syn
        self.tcp_option = tcp_option
        self.mss = mss
        self.datestart = datestart
        self.datestop = datestop
        self.timestart = timestart
        self.timestop = timestop
        self.monthdays = monthdays
        self.weekdays = weekdays
        self.utc = utc
        self.localtz = localtz
        self.tos = tos
        self.ttl_eq = ttl_eq
        self.ttl_gt = ttl_gt
        self.ttl_lt = ttl_lt
        self.u32 = u32
        self.set_class = set_class
        self.new = new
        self.hashmode = hashmode
        self.clustermac = clustermac
        self.total_nodes = total_nodes
        self.local_nodes = local_nodes
        self.hash_init = hash_init
        self.set_xmark = set_xmark
        self.save_mark = save_mark
        self.restore_mark = restore_mark
        self.and_mark = and_mark
        self.or_mark = or_mark
        self.xor_mark = xor_mark
        self.to_destination = to_destination
        self.random = random
        self.set_dscp = set_dscp
        self.set_dscp_class = set_dscp_class
        self.ecn_tcp_remove = ecn_tcp_remove
        self.log_level = log_level
        self.log_prefix = log_prefix
        self.log_tcp_sequence = log_tcp_sequence
        self.log_tcp_options = log_tcp_options
        self.log_ip_options = log_ip_options
        self.log_uid = log_uid
        self.to_ports = to_ports
        self.nflog_group = nflog_group
        self.nflog_prefix = nflog_prefix
        self.nflog_range = nflog_range
        self.nflog_threshold = nflog_threshold
        self.queue_num = queue_num
        self.ratetest_name = ratetest_name
        self.ratetest_interval = ratetest_interval
        self.ratetest_ewmalog = ratetest_ewmalog
        self.reject_with = reject_with
        self.nodst = nodst
        self.selectx = selectx
        self.add_set = add_set
        self.del_set = del_set
        self.to_source = to_source
        self.set_mss = set_mss
        self.clamp_mss_to_pmtu = clamp_mss_to_pmtu
        self.strip_options = strip_options
        self.set_tos = set_tos
        self.and_tos = and_tos
        self.or_tos = or_tos
        self.xor_tos = xor_tos
        self.ttl_set = ttl_set
        self.ttl_dec = ttl_dec
        self.ttl_inc = ttl_inc
        self.ulog_nlgroup = ulog_nlgroup
        self.ulog_prefix = ulog_prefix
        self.ulog_cprange = ulog_cprange
        self.ulog_qthreshold = ulog_qthreshold
        self.ipp2p = ipp2p
        self.edonkey = edonkey
        self.kazaa = kazaa
        self.gnutella = gnutella
        self.bittorrent = bittorrent
        self.apple = apple
        self.winmx = winmx
        self.directconnect = directconnect
        self.soulseek = soulseek
        self.ares = ares
        self.status_params = status_params

    def __repr__(self):
        return self.__str__()

    def update(self, *dictionary):
        """Just an alias for self.__dict__.update()"""
        self.__dict__.update(*dictionary)

    def iptables_line(self, output_template=rule_template):
        """Returns the iptables command line that can be used to apply the rule.
        Takes 'output_template' as an argument which is the Mako template to be
        used when producing the iptables line.

        Why use a different template?  The default template has logic in it to
        cover just about every permutation of an iptables rule.  If you know ahead
        of time that your rules will only contain a specific subset of command
        line switches you can use a template with less permutations to speed up
        the process considerably."""
        output = Template(output_template)
        # Now reduce it to one line
        for line in output.render(rules=[self]).splitlines():
            if line.strip().rstrip('\n'):
                return " ".join(line.split()) # Remove unnecessary whitespace

    def __str__(self):
        return self.iptables_line()

class Chain(object):
    """Class to store IPTables chains"""
    def __init__(self,
        name,
        table,
        policy=None,
        references=None,
        rules=None,
        packets=None,
        bytes=None
    ):
        self.name = name
        self.policy = policy
        self.table = table # i.e. 'nat', 'filter', 'mangle', or 'raw'
        self.packets = packets
        self.bytes = bytes
        if not rules:
            rules = []
        self.rules = rules # For storing Rule() objects
        if references:
            self.references = references # This is the number of times this chain is listed as a target in other chains
            self.custom = True
        else:
            self.custom = False

    def __repr__(self):
        try:
            return str(self.rules)
        except:
            return str([])

    def __str__(self):
        try:
            return str(self.rules)
        except:
            return str([])

    def append_rule(self, rule):
        """Appends a Rule() object to self.rules"""
        self.rules.append(rule)

    def __iter__(self):
        return self.rules.__iter__()

    def __len__(self):
        return len(self.rules)

    def __contains__(self, val):
        for item in self.rules:
            if item == val:
                return True
        else:
            return False

    def __getitem__(self, val):
        """This lets you access Chain objects as a dictionary or a list."""
        try:
            return self.__dict__[val]
        except:
            try:
                return self.rules[val] # val is an index
            except:
                return None

    #def __setitem__(self, key, value):
        #"""This lets you set/assign attributes to Chain objects like a dictionary."""
        #self.__dict__[key] = value

class Table(object):
    """Class to store Table() objects (one of 'filter', 'nat', 'mangle' or 'raw')"""
    def __init__(self, name, chains=[]):
        """Initialize the table, 'name'"""
        self.name = name
        self.chains = chains # Should be a dict of chains belonging to this table

    def __str__(self):
        return str(chains)

    def append_chain(self, chain):
        """Appends the given Chain object to self.chains"""
        self.chains.append(chain)

class IPTables(ConfigFile):
    """Class to read and store iptables rules"""
    def __init__(self, filepath='/etc/firewall.user'):
        """Read /etc/firewall.user into a list of Chain() objects each with it's own options."""
        #TODO: Make this preserve comments in config files
        self.filepath = filepath
        self.tables = {'filter': Table('filter'), 'nat': Table('nat'), 'mangle': Table('mangle'), 'raw': Table('raw')}
        self.parse_config_active()
        self.rules = [] # Added this for quick access while testing

    def parse_config_active(self):
        """Parses the output of 'iptables -t <tables> -L -n -v -x --line-numbers'
        and creates/assigns Chain objects to self.chains."""
        tables = ['filter', 'nat', 'mangle', 'raw']
        for table in tables:
            table_obj = Table(table)
            table_chains = {}
            command = "/usr/bin/env iptables -t %s -L -n -v -x --line-numbers" % table
            p = subprocess.Popen(
                command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
            stdout, stderr = p.communicate()
            for line in stdout.splitlines():
                if line.startswith('Chain'):
                    if 'policy' in line: # Vanilla chain
                        line = line.split()
                        chain_name = line[1]
                        policy = line[3]
                        packets = line[4]
                        bytes = line[6]
                        chain_obj = Chain(
                            chain_name,
                            table,
                            policy=policy,
                            packets=packets,
                            bytes=bytes
                        )
                        table_obj.append_chain(chain_obj)
                        table_chains.update({chain_name: chain_obj})
                    elif 'references' in line: # Custom chain
                        line = line.split()
                        chain_name = line[1]
                        references = line[2].rstrip('(') # turn '(1' into '1'
                        chain_obj = Chain(chain_name, table, references=references)
                        table_obj.append_chain(chain_obj)
                        table_chains.update({chain_name: chain_obj})
                # Lines that start with 'num' are just headers...
                elif not line.startswith('num') and len(line) > 1:
                    line = line.split()
                    ruleno = line[0]
                    packets = line[1]
                    bytes = line[2]
                    target = line[3]
                    protocol = line[4]
                    target_opts = line[5]
                    in_interface = line[6]
                    out_interface = line[7]
                    source = line[8]
                    destination = line[9]
                    status_params = None # What do you call the 11th column? It has no header
                    match = None
                    state = None
                    ctstate = None
                    ipp2p = None
                    reject_with = None
                    tcp_flags = None
                    limit = None
                    limit_burst = None
                    destination_port = None
                    source_port = None
                    try:
                        status_params = " ".join(line[10:])
                        #print "status_params: %s" % status_params
                        if not status_params:
                            match = None # as oppsed to ''
                        else:
                            if status_params.startswith('tcp'): # TODO:  Figure out all the permutations of 'tcp' and 'udp'
                                if status_params.split()[1].startswith('flags'): # TCP Flags rule
                                    # Example:  tcp flags:FIN,SYN,RST,ACK/SYN limit: avg 25/sec burst 50
                                    tcp_flags = match.split()[1].split(':')[1]
                                elif status_params.split()[1].startswith('dpt'): # Destination port
                                    destination_port = status_params.split()[1].split(':')[1]
                                elif status_params.split()[1].startswith('spt'): # Source port
                                    source_port = status_params.split()[1].split(':')[1]
                            if 'limit:' in status_params:
                                limit = status_params.split('limit: avg ')[1].split()[0]
                                if 'burst' in status_params:
                                    limit_burst = status_params.split('burst ')[1]
                            if 'reject-with' in status_params:
                                reject_with = status_params.split('reject-with ')[1].split()[0]
                            else:
                                match = status_params.split()[0] # Example: 'state RELATED,ESTABLISHED'
                                if match == 'state':
                                    state = match.split()[1:]
                                elif match == 'ctstate':
                                    ctstate = match.split()[1:]
                                elif match == 'ipp2p': # TODO: Add all the ipp2p options
                                    ipp2p = True
                    except:
                        match = None
                    rule_obj = Rule(
                        table=table_obj,
                        chain=chain_name,
                        status_params=status_params,
                        match=match,
                        target=target,
                        protocol=protocol,
                        target_opts=target_opts,
                        source=source,
                        destination=destination,
                        in_interface=in_interface,
                        out_interface=out_interface,
                        packets=packets,
                        bytes=bytes,
                        ruleno=ruleno,
                        state=state,
                        ctstate=ctstate,
                        ipp2p=ipp2p,
                        reject_with=reject_with,
                        tcp_flags=tcp_flags,
                        limit=limit,
                        limit_burst=limit_burst,
                        destination_port=destination_port,
                        source_port=source_port
                    )
                    chain_obj.rules.append(rule_obj)
                    #print "chain_obj rules: %s" % chain_obj.rules
            self.tables.update({table: table_chains})

    def reality_check(self):
        """Compares the configuration in self.filepath against the iptables rules set on the running host."""
        # TODO:  Write this
        pass

    def parse_config_file(self):
        """Reads self.filepath and returns a list of Chain objects (each containing Rule objects) representing the file."""
        for line in open(self.filepath):
            if 'iptables' in line and not line.startswith('#'): # Skip comments
                line = line.rstrip('\n')
                args = line.split('iptables')[1:] # Gets everything after 'iptables' in the line
                args = [args[0].strip()]
                args = args[0].split() # OptionParser wants arguments to be broken up
                parsed = parse_iptables_args(args)
                if parsed.table not in ['filter', 'mangle', 'raw', 'nat']:
                    raise RuleParsingException('Table %s does not exist!' % parsed.table)
                rule = Rule(table=parsed.table, chain=parsed.chain)
                for key, value in parsed.__dict__.items():
                    if value:
                        rule.update({key: value})
                        #print "key: %s\t value: %s" % (key, value)
                self.rules.append(rule)
                if parsed.chain in self.tables[parsed.table].keys():
                    self.tables[parsed.table][parsed.chain].rules.append(rule)

if __name__ == '__main__':
    """If we're run directly on the command line, print out all the firewall rules we find
    in the given filename from sys.argv[1]"""
    import sys

    if len(sys.argv) < 2:
        print "syntax: %s filename" % sys.argv[0]
        sys.exit(1)

    filename = sys.argv[1]
    for line in open(filename):
        if 'iptables' in line and not line.startswith('#'):
            line = line.rstrip('\n')
            args = line.split('iptables')[1:] # Gets everything after 'iptables' in the line
            args = [args[0].strip()]
            args = args[0].split()
            parsed = parse_iptables_args(args)
            for key, value in parsed.__dict__.items():
                if value:
                    print "key: %s\t value: %s" % (key, value)