#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       l2sh.py
#
#       Copyright 2009 Dan McDougall <YouKnowWho@YouKnowWhat.com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; Version 3 of the License
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, the license can be downloaded here:
#
#       http://www.gnu.org/licenses/gpl.html

# Meta
__version__ = '0.4'
__license__ = "GNU General Public License (GPL) Version 3"
__version_info__ = (0, 4)
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

# NOTE:  This l2 telnet client/server work but not as well as they should.  Here's the important TODO list:
#
#   1) The server should support more than one client connection per interface at a time.
#   2) The client should intercept all keyboard input (things like control-C) and send that to the server (as opposed to exiting).
#   3) Full-screen apps like vi don't work properly--this needs to be fixed.
#
# Other than that, it should be good enough to get into your router to run "ifconfig", "route" and "iptables" commands
# ...to get back in via ssh =)

import sys, os, thread, getpass, select
from datetime import datetime, timedelta
from commands import getstatusoutput
from time import sleep
import struct
from socket import *
from optparse import OptionParser

# Globals
PROTO = 0x55aa
STDIN, STDOUT, STDERR = 0, 1, 2
L2SESSION_COUNT = 0
TIMEOUT = 300 # seconds

def counter():
    """Adds to the count every time it is called."""
    global L2SESSION_COUNT
    L2SESSION_COUNT += 1
    return L2SESSION_COUNT

# TODO:  Get this thread version of the server working
#class L2ShellThread(threading.Thread):
    #"""
    #This thread spawns a new shell as the supplied username and controls the
    #session with the client.  Optionally, a different shell and arguments may be
    #supplied as 'sh', and 'args' respectively.
    #"""
    #def __init__(self, username, dstAddr, sock, sh='/bin/sh', args=['-l']):
        #threading.Thread.__init__(self, name="L2ShellThread-%s-%s" % (username, datetime.now().isoformat()))
        #self.quitting = False
        #self.username = username
        #self.dstAddr = dstAddr
        #self.sock = sock
        #self.sh = sh
        #self.args = args

    #def quit(self):
        #global L2SESSION_COUNT
        #self.quitting = True
        #L2SESSION_COUNT -= 1

    #def l2sh_shell(self, username, dstAddr):
        #"""Spawns a new interactive shell as the given username and communicates
        #with the client over the given destination address and socket."""
        #ifName, ifProto, pktType, hwType, hwAddr = self.sock.getsockname()
        #self.srcAddr = hwAddr
        #self.pid, self.local = pty.fork()
        #self.remote = self.sock.fileno()

        #if self.pid == 0:
            #attrs = tty.tcgetattr(1)
            #attrs[3] = attrs[3] & ~tty.ICANON & ~tty.ECHO
            #attrs[6][tty.VMIN] = 1
            #attrs[6][tty.VTIME] = 0
            #tty.tcsetattr(1, tty.TCSANOW, attrs)
            ## Spawn the shell as the authenticated user
            #os.seteuid(get_uid(self.username))
            #os.execv(self.sh, [self.sh, self.args])
        #else:
            #while not self.quitting:
                #toRead = select.select([self.local, self.remote], [], [], 0.1)[0]
                #print toRead
                #print "pid: %s" % self.pid
                #print "self.local: %s" % `self.local`
                #if self.local in toRead:
                    #data = os.read(self.local, 2048)
                    #txFrame = struct.pack("!6s6sh", self.dstAddr, self.srcAddr, PROTO) + ')(' + data
                    #self.sock.send(txFrame)
                    #continue
                #if self.remote in toRead:
                    #frame = os.read(self.remote, 2048)
                    #dstAddr, srcAddr, proto = struct.unpack("!6s6sh", frame[:14])
                    #data = frame[14:]
                    #if data.startswith('()'):
                        #os.write(self.local, data.lstrip('()'))
                    #continue
            #os.kill(self.pid, signal.SIGKILL)

    #def run(self):
        #self.l2sh_shell(self.username, self.dstAddr)

#def start_l2sh_session(username, dstAddr, sock):
    #"""Starts a new L2ShellThread() session."""
    #l2_shell_thread = L2ShellThread(username, dstAddr, sock)
    #l2_shell_thread.start()
    #return l2_shell_thread.name

#def stop_l2sh_session(threadname):
    #"""Stops the given session (using its name)"""
    #for t in threading.enumerate():
        #if t.getName() == threadname:
            #t.quit()

def interfaces():
    """Returns a list of all non-alias, UP interfaces on this host"""
    out_list = []
    exit_code, output = getstatusoutput('ip -o link')
    for line in output.splitlines():
        if 'link/ether' in line and ',UP' in line:
            out_list.append(line.split()[1].rstrip(':'))
    return out_list

def setup_socket(interface):
    """Returns a socket object bound to the given interface (e.g. 'eth0').
    The socket is created as such: socket(AF_PACKET, SOCK_RAW, PROTO)
    Note: By default PROTO is set to 0x55aa"""
    s = socket(AF_PACKET, SOCK_RAW, PROTO)
    s.bind((interface, PROTO))
    return s

def get_uid(username):
    """Returns the UID of the given user"""
    from pwd import getpwnam
    return getpwnam(username)[2]

def check_credentials(user, password):
    """Checks a given user and password against /etc/shadow (or /etc/passwd if /etc/shadow doesn't exist).
    Returns None on success and an error message on failure."""
    from crypt import crypt
    import re
    shadow_hash = ''
    salt_regex = re.compile(r'\$.*\$.*\$')
    if os.path.exists('/etc/shadow'):
        password_file = '/etc/shadow'
    else:
        password_file = '/etc/passwd'
    shadow = open(password_file, 'r').readlines()
    for line in shadow:
        cols = line.split(':')
        if cols[0] == user:
            shadow_hash = cols[1]
    if salt_regex.match(shadow_hash):
        salt = salt_regex.match(shadow_hash).group()
        hashed_pass = crypt(password, salt)
        if hashed_pass == shadow_hash:
            return None
    # Now check the case where OpenWRT was just started up for the first time and there's no root password set yet...
    # If the root password hasn't been set the telnet daemon will be running with /bin/login.sh (i.e. no-password login)
    retcode = getstatusoutput('ps aux | grep -v grep | grep "telnetd -l /bin/login.sh"')[0]
    re.purge() # This empties the regex cache, freeing memory
    if retcode == 0: # A return code of 0 means grep actually grepped something (as opposed to no output)
        return None # Let them through
    else:
        return "Invalid user and/or password"

def start_l2sh_server(sh, arg, sock):
    ifName, ifProto, pktType, hwType, hwAddr = sock.getsockname()
    print "Started l2sh listener on %s" % `ifName`
    srcAddr = hwAddr
    while True:
        rxFrame = sock.recv(2048)
        dstAddr, srcAddr, proto = struct.unpack("!6s6sh",rxFrame[:14])
        ethData = rxFrame[14:].rstrip('\x00') # Remove trailing null bytes
        quitting = False
        if ethData.startswith('CONNECT'):
            username = ethData.split()[1]
            password = ethData.split()[2]
            result = check_credentials(username, password)
            if not result:
                # TODO:  Change this to use logging
                print "%s authenticated successfully" % username
                alive = datetime.now()
                #pid, local = pty.fork()
                (local, slave) = os.openpty()
                child=os.fork()
                if not child:
                    os.setsid()
                    os.dup2(slave,sys.stdin.fileno())
                    os.dup2(slave,sys.stdout.fileno())
                    os.dup2(slave,sys.stderr.fileno())
                    os.close(slave)
                    os.close(local)
                    os.seteuid(get_uid(username))
                    os.execv(sh, arg)
                    sys.exit(1)
                os.close(slave)
                remote = sock.fileno()
                #if pid:
                    #attrs = tty.tcgetattr(1)
                    #attrs[3] = attrs[3] & ~tty.ICANON & ~tty.ECHO
                    #attrs[6][tty.VMIN] = 1
                    #attrs[6][tty.VTIME] = 0
                    #tty.tcsetattr(1, tty.TCSANOW, attrs)
                    ## Spawn the shell as the authenticated user
                    #os.seteuid(get_uid(username))
                    #os.execv(sh, arg)
                #else:
                while not quitting:
                    toRead = select.select([local, remote], [], [], 0.1)[0]
                    if datetime.now() > alive + timedelta(seconds=TIMEOUT):
                        print "l2sh: client timed out"
                        quitting = True
                    if local in toRead:
                        data = os.read(local, 1024)
                        #print "local shell: %s" % data
                        txFrame = struct.pack("!6s6sh",srcAddr,dstAddr,PROTO) + ')(' + data
                        try:
                            sock.send(txFrame)
                        except Exception as e:
                            print e # Annoyingly, we occasionally get "[Errno 90] Message too long" (haven't figured it out yet)
                            # Oddly, if I ignore it the frame appears to get sent fine anyway
                        continue
                    if remote in toRead:
                        frame = os.read(remote, 1024)
                        dstAddr, srcAddr, proto = struct.unpack("!6s6sh", frame[:14])
                        data = frame[14:].rstrip('\x00') # Remove trailing null bytes
                        if data.startswith('()'):
                            #print "data: %s" % `data.lstrip('()')`
                            alive = datetime.now()
                            os.write(local, data.lstrip('()'))
                        continue
                os.close(local)
                #os.kill(pid, signal.SIGKILL)
            else:
                txFrame = struct.pack("!6s6sh",dstAddr,srcAddr,PROTO) + ')(Login failed!'
                sock.send(txFrame)
        sleep(0.1)

def start_l2sh_client(dstAddr, username, password, sock):
    local = STDIN
    remote = sock.fileno()
    ifName, ifProto, pktType, hwType, hwAddr = sock.getsockname()
    srcAddr = hwAddr
    # This constructs the proper connect packet (with username and password):
    ethData = "CONNECT %s %s" % (username, password)
    txFrame = struct.pack("!6s6sh",dstAddr,srcAddr,PROTO) + ethData
    sock.send(txFrame)
    # TODO:  Write a piece here that makes sure we authenticated successfully and re-asks for a password if not
    #        ...so the user doesn't have to control-C and try again
    while True:
        toRead = select.select([local, remote], [], [], 0.1)[0]
        if local in toRead:
            data = os.read(local, 1024)
            txFrame = struct.pack("!6s6sh",dstAddr,srcAddr,PROTO) + '()' + data
            sock.send(txFrame)
            continue
        if remote in toRead:
            frame = os.read(remote, 1024)
            server_dstAddr, server_srcAddr, proto = struct.unpack("!6s6sh", frame[:14])
            data = frame[14:].rstrip('\x00') # Remove trailing null bytes
            #print "data: %s" % data
            # Only write the received data if it is in the proper format:
            if data.startswith(')('):
                os.write(STDOUT, data.lstrip(')('))
            continue

# TODO:  Write some more sophisticated negotiation and sessioning:
#           * The server should send the client a random string that must be prepended before all subsequent commands
#           * if the server receives a packet without that string it should ignore it.
#           * Add in some rudimentary checksumming to the packets
# TODO;  Wrap this in a GUI like speedtestclient.py

# LONG TERM TODO:  Get this using encryption like ssh.  In other words, I'd like this to be l2ssh which would be 900x better.
#                  ...I'd work on this now but I only have a week left before the RouterStation contest ends!
#                  Internet at-large, please forgive this security faux pas for the time being.  I promise to provide the
#                  world with a robust and working layer 2 SSH implementation as soon as I figure out how!

if __name__ == '__main__':
    usage = 'usage: %prog [options] [MAC Address]'
    parser = OptionParser(usage=usage)
    parser.add_option("-c", "--client", action="store_true", dest="client", default=True, help="Run in client mode connecting to the given host (default).")
    parser.add_option("-i", "--interface", dest="interface", default='all', help="Bind to the given interface(s) (e.g. '-i eth0,wlan0') or 'all' to bind on all UP interfaces.", metavar="<interface>")
    parser.add_option("-s", "--server", action="store_true", dest="server", default=False, help="Run in server mode.")
    parser.add_option("-u", "--username", dest="username", default=None, help="The username to be used when connecting (uses your logged-in username by default).", metavar="<username>")
    parser.add_option("-P", "--password", dest="password", default=None, help="The password to be used when connecting (will ask if not provided--recommended)", metavar="<password>")
    (options, args) = parser.parse_args()

    # Check to make sure we were passed at least one command line argument
    try:
        sys.argv[1]
    except:
        print "\nError:  At a minimum you must supply -i <interface> '<MAC address>' (client mode connecting to <MAC address>) or '-s -i <interface>' (server mode)"
        parser.print_help()
        sys.exit(2)

    if not os.geteuid()==0:
        sys.exit("Error:  You must become root before running this program.\n")

    if args:
        dstAddr = args[0]
        # We need to convert human-formatted MAC addresses to this format:  \x00\x00\x00\x00\x00\x00
        if ':' in dstAddr: # MAC address is like so: 00:00:00:00:00:00
            base16_list = []
            hex_list = []
            # First convert to base16...
            for i in dstAddr.split(':'):
                base16_list.append(int(i, 16))
            # Next convert to hex...
            for i in base16_list:
                hex_list.append(chr(i))
            # Lastly we join it all back to a string and clean up:
            dstAddr = "".join(hex_list)
            del base16_list
            del hex_list
            #print dstAddr

    if ',' in options.interface:
        # Convert things like, 'wlan0,eth0' into a list
        options.interface = options.interface.split(',')

    if not options.server:
        if not options.username:
            options.username = raw_input('Username: ')
        if not options.password:
            options.password = getpass.getpass('Password: ')

    try:
        if options.server: # Run in server mode
            print 'serving on %s (layer 2)...' % `options.interface`
            if isinstance(options.interface, list):
                for i in options.interface:
                    sock = setup_socket(i)
                    thread.start_new_thread (start_l2sh_server, ('/bin/sh', ['/bin/sh'], sock))
            elif options.interface == 'all':
                for i in interfaces():
                    sock = setup_socket(i)
                    thread.start_new_thread (start_l2sh_server, ('/bin/sh', ['/bin/sh'], sock))
            else:
                sock = setup_socket(options.interface)
                thread.start_new_thread (start_l2sh_server, ('/bin/sh', ['/bin/sh'], sock))
            while True:
                sleep(1)

        elif options.client: # Run in client mode
            print "Press control-C to exit (yes, this means you can't use commands that require a SIGINT--working on it!)."
            sock = setup_socket(options.interface)
            start_l2sh_client(dstAddr, options.username, options.password, sock)

    except KeyboardInterrupt:
        sys.exit(0)