#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os, sys, tarfile, datetime
from commands import getstatusoutput
from platform import uname
from optparse import OptionParser

def backup_etc(backup_dir, rotate):
    """Saves a compressed tarball containing the contents of /etc to 'backup_dir'.
    Keeps 'rotate' (int) number of increments and removes older backups.
    'backup_dir' should be something like '/mnt/sda1' (directory where backup
    files will wind up)."""

    # First we make the backup file
    now = datetime.datetime.now()
    hostname = uname()[1]
    filepath = "%s/pyci_backup_%s_etc-%s.tar.gz" % (backup_dir, hostname, now.strftime('%Y%m%d_%H%M%S'))
    tar = tarfile.open(filepath, 'w:gz')
    os.chdir('/')
    tar.add('/etc') # Add everything in /etc to our tarball
    tar.close()

    # Now we perform the rotation (remove the oldest of 'rotate' series)
    retcode, output = getstatusoutput('ls -lt %s/pyci_backup*' % backup_dir)
    lines = output.splitlines()
    num_lines = len(lines)
    if num_lines > rotate:
        oldest_file = lines[num_lines - 1].split()[7]
        oldest_file_path = '%s' % oldest_file
        os.remove(oldest_file_path)

if __name__ == '__main__':
    usage = 'usage: %prog [options]'
    parser = OptionParser(usage=usage)
    parser.add_option("-d", "--dir", dest="backup_dir", help="Backup files will be saved at this path (e.g. '/mnt/sda1')", metavar="<path>")
    parser.add_option("-r", "--rotate", dest="rotate", default=7, help="The number of backup files to keep (default: 7).", metavar="<username>")
    (options, args) = parser.parse_args()

    if not isinstance(options.rotate, int):
        options.rotate = int(options.rotate)

    # Check to make sure we were passed at least one command line argument
    try:
        sys.argv[1]
    except:
        print "\nError:  At a minimum you must supply -d <path to backup directory>"
        parser.print_help()
        sys.exit(2)

    if not os.geteuid()==0:
        sys.exit("Error:  You must become root before running this program.\n")

    # Do the backup
    try:
        backup_etc(options.backup_dir, options.rotate)
    except Exception as e:
        print "Got an exception trying to backup /etc: %s" % e