#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Description: Quagga config file parser

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

import os, re, gettext
from odict import odict

locale_dir = os.path.join(os.getcwd(), 'i18n')
def get_translation():
    try:
        trans = gettext.translation('PyCI', locale_dir)
    except:
        trans = gettext.NullTranslations()
    return trans
trans = get_translation()
_ = lambda s: trans.ugettext(s).encode('utf-8')

# Notes:
#   We've got these special kinds of configuration options:
#       <1-###>, (option1|option2|option3|...), and 'name'/'NAME', a.b.c.d, A.B.C.D/M, WORD, X:X::X:X, IPV4_PREFIX, IPV6_PREFIX, peer, description, access-list, prefix-list, path, interval, ifname, password, minute, second, time, hostname, address/prefix, {permit|deny}, line, community, extcommunity, AUTH_KEY, number, len, key-chain, version, string, network, gateway, distance, facility, level, ipv4_addr

# Regular expression matching for each special type:
#ip_regex = re.compile(r"\b(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\b")
#ip_subnet_regex = re.compile(r"\b(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\/[0-9][0-9]?\b")
#ipv6_regex = re.compile('^\s*((([0-9A-Fa-f]{1,4}:){7}(([0-9A-Fa-f]{1,4})|:))|(([0-9A-Fa-f]{1,4}:){6}(:|((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})|(:[0-9A-Fa-f]{1,4})))|(([0-9A-Fa-f]{1,4}:){5}((:((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})?)|((:[0-9A-Fa-f]{1,4}){1,2})))|(([0-9A-Fa-f]{1,4}:){4}(:[0-9A-Fa-f]{1,4}){0,1}((:((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})?)|((:[0-9A-Fa-f]{1,4}){1,2})))|(([0-9A-Fa-f]{1,4}:){3}(:[0-9A-Fa-f]{1,4}){0,2}((:((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})?)|((:[0-9A-Fa-f]{1,4}){1,2})))|(([0-9A-Fa-f]{1,4}:){2}(:[0-9A-Fa-f]{1,4}){0,3}((:((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})?)|((:[0-9A-Fa-f]{1,4}){1,2})))|(([0-9A-Fa-f]{1,4}:)(:[0-9A-Fa-f]{1,4}){0,4}((:((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})?)|((:[0-9A-Fa-f]{1,4}){1,2})))|(:(:[0-9A-Fa-f]{1,4}){0,5}((:((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})?)|((:[0-9A-Fa-f]{1,4}){1,2})))|(((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})))(%.+)?\s*$')
#ipv6_subnet_regex = re.compile('^\s*((([0-9A-Fa-f]{1,4}:){7}(([0-9A-Fa-f]{1,4})|:))|(([0-9A-Fa-f]{1,4}:){6}(:|((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})|(:[0-9A-Fa-f]{1,4})))|(([0-9A-Fa-f]{1,4}:){5}((:((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})?)|((:[0-9A-Fa-f]{1,4}){1,2})))|(([0-9A-Fa-f]{1,4}:){4}(:[0-9A-Fa-f]{1,4}){0,1}((:((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})?)|((:[0-9A-Fa-f]{1,4}){1,2})))|(([0-9A-Fa-f]{1,4}:){3}(:[0-9A-Fa-f]{1,4}){0,2}((:((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})?)|((:[0-9A-Fa-f]{1,4}){1,2})))|(([0-9A-Fa-f]{1,4}:){2}(:[0-9A-Fa-f]{1,4}){0,3}((:((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})?)|((:[0-9A-Fa-f]{1,4}){1,2})))|(([0-9A-Fa-f]{1,4}:)(:[0-9A-Fa-f]{1,4}){0,4}((:((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})?)|((:[0-9A-Fa-f]{1,4}){1,2})))|(:(:[0-9A-Fa-f]{1,4}){0,5}((:((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})?)|((:[0-9A-Fa-f]{1,4}){1,2})))|(((25[0-5]|2[0-4]\d|[01]?\d{1,2})(\.(25[0-5]|2[0-4]\d|[01]?\d{1,2})){3})))(%.+)?\/[0-9][0-9]?[0-9]?\s*$')

class QuaggaConfig(object):
    """Class object to store Quagga configuration files"""

    validation_file = "metadata/quagga"
    # These are for figuring out validation and option enumeration
    # TODO:  Finish validation or let the quagga daemons try to execute with these files (they return their own validation errors)
    ipv4_types = ['a.b.c.d', 'A.B.C.D', 'ipv4_addr', 'gateway']
    ipv4_subnet_types = ['a.b.c.d/m', 'A.B.C.D/M', 'network', 'IPV4_PREFIX']
    ipv6_types = ['address/prefix', 'ipv6_addr']
    word_types = ['name', 'NAME', 'WORD', 'word', 'hostname']
    number_types = ['number', 'NUMBER']
    range_regex = re.compile('\<[0-9]\-[0-9]*\>')
    hostname_regex = re.compile("^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z]|[A-Za-z][A-Za-z0-9\-]*[A-Za-z0-9])$")

    # Help texts taken from the official Quagga documentation here: http://www.quagga.net/docs/docs-info.php
    help = {
        'comment': _("Lines starting with a bang (!) will be ignored."),
        'hostname': _('Set hostname of the router.'),
        'password': _("Set password for vty interface. If there is no password, a vty won't accept connections."),
        'enable password': _("Set enable password (which is used to enter 'enable' mode on a vty"),
        'log syslog': _("<dl><dt>log syslog <i>level</i></dt><dd>Enable logging output to syslog.  If a logging <i>level</i> is not present, the default logging level will be used.</dd></dl>"),
        'log stdout': _("<dl><dt>log stdout <i>level</i></dt><dd>Enable logging output to stdout.  If a logging <i>level</i> is not present, the default logging level will be used.</dd></dl>"),
        'log file': _("<dl><dt>log file <i>filename level</i></dt><dd>If you want to log into a file, please specify filename as in this example:<br /><code>log file /var/log/quagga/bgpd.log informational</code><br />If a logging <i>level</i> is not present, the default logging level will be used.</dd></dl>"),
        'log facility': _("<dl><dt>log facility <i>facility</i></dt><dd>This command changes the facility used in syslog messages. The default facility is daemon.</dd></dl>"),
        'log record-priority': _("<dl><dt>log record-priority</dt><dd>To include the severity in all messages logged to a file, to stdout, or to a terminal monitor (i.e. anything except syslog), use the log record-priority global configuration command.</dd></dl>"),
        'service password-encryption': _("<dl><dt>service password-encryption</dt><dd>Encrypt password (that's all the Quagga docs have to say on this!).</dd></dl>"),
        'service advanced-vty': _("<dl><dt>service advanced-vty</dt><dd>Enable advanced mode VTY (another great explanation from the Quagga docs).</dd></dl>"),
        'service terminal-length': _("<dl><dt>service terminal-length <i>&lt;0-512&gt;</i></dt><dd>Set system wide line configuration. This configuration command applies to all VTY interfaces.</dd></dl>"),
        'banner': _("<dl><dt>banner motd default</dt><dd>Set default motd string.</dd></dt>"),
        'exec-timeout': _("""<dl><dt>exec-timeout <i>minute second</i></dt><dd>Set VTY connection timeout value. When only one argument is specified it is used for timeout value in minutes. Optional second argument is used for timeout value in seconds. Default timeout value is 10 minutes. When timeout value is zero, it means no timeout.</dd></dt>"""),
        'access-class': _("<dl><dt>access-class <i>access-list</i></dt><dd>Restrict vty connections with an access list.</dd></dl>"),
        # TODO:  Finish adding these.
    }

    paired_commands = ['enable', 'router', 'log', 'service', 'set', 'bgp', 'ip', 'ip6', 'ipv6', 'auto-cost', 'clear', 'debug', 'default-information', 'dump', 'line', 'match', 'max-metric', 'on-match', 'ospf', 'redistribute', 'timers']

    def __init__(self, filepath):
        """Parse the config file at 'filepath' and set up our variables"""
        self.filepath = filepath
        self.config = self.parse_config(self.filepath)
        for pair in self.config:
            if pair[0] == 'password':
                self.password = pair[1]
        for pair in self.config:
            if pair[0] == 'enable password':
                self.enable_password = pair[1]

    def parse_config(self, filepath):
        """
        Parses the given quagga configuration file and returns a list of configuration pairs like so:
        [
            ('hostname', 'foo'),
            ('password', 'insertpasswordhere'),
            ('enable password', 'insertpasswordhere'),
            ('comment', '! interfaces\n'),
            ('interface', 'lo'),
            ('interface', 'eth0'),
            ('interface', 'vlan0'),
            ('interface', 'vlan1'),
            ('ip', 'route 10.10.10.0/24 Null0 255'),
            ('line', 'vty'),
            ('access-class', 'vty')
        ]
        """
        out_list = []
        for line in open(filepath):
            if line.startswith('!'): # Quagga comments start with bangs
                value = line.lstrip('!').rstrip('\n').strip()
                out_list.append(('comment', value))
            else:
                command = line.split()[0]
                if command in self.paired_commands: # Special case where the command is two words
                    command = "%s %s" % (command, line.split()[1])
                    value = " ".join(line.split()[2:])
                elif command == 'no': # Lots of commands can start with 'no'
                    command = command + " " + line.split()[0]
                    value = " ".join(line.split()[2:])
                else:
                    value = " ".join(line.split()[1:])
                out_list.append((command, value))
        return out_list

    def valid_commands(self):
        """Returns a list of valid quagga commands (just the command, no options)"""
        out_list = []
        for line in open(self.validation_file):
            if line.startswith('#'):
                pass # Ignore comments
            else:
                command = line.split()[0]
                if command in self.paired_commands: # Special case where the command is two words
                    command = "%s %s" % (command, line.split()[1])
                if command not in out_list:
                    out_list.append(command)
        return out_list

    def valid_options(self, command):
        """Returns a list of valid options that can be passed to 'command'"""
        out_list = []
        for line in open(self.validation_file):
            if line.startswith('#'):
                pass # Ignore comments
            elif line:
                if len(line.split()) == 1:
                    pass # No options for this command
                elif line.split()[0] == command:
                    if line.split()[1] not in out_list:
                        out_list.append(line.split()[1])
        return out_list

    def validate_config(self, filepath):
        """
        Validates the commands in the config file found at 'filepath'.
        Returns None if no errors are found.
        Returns an error list with the following format if errors are found:
            [[line_number, error_message, line], ...]
        Example:
            >>> f = open('/tmp/zebra.conf', 'w')
            >>> f.write('foo start')
            >>> f.close()
            >>> validate_config('/tmp/zebra.conf')
            [[123, "foo is not a valid Quagga command", "foo start"]]
        """
        #import ipaddr
        bad_lines = []
        for lineno, line in enumerate(open(filepath)):
            if line.startswith('!'): # Quagga comments start with bangs
                pass # Ignore comments
            else:
                # Validate commands
                command = line.split()[0]
                if command not in valid_commands():
                    error = "%s is not a valid Quagga command" % command
                    bad_lines.append([lineno, error, line])
                # TODO: Validate options
                # valid_opts = valid_options(command)
        if bad_lines:
            return bad_lines

telnet_connection = None
prompt = None

def vtysh_command(command, password, enable=False, enable_password=None, host="127.0.0.1", port=None, daemon=None, timeout=2):
    """
    Executes the given command via Quagga's vtysh (telnet) interface on the given host/port
    and returns the result as a string.
    if 'daemon' is set to the name of one of the Quagga daemons (e.g. 'zebra', 'bgpd', etc)
    the default port for that daemon will be used.
    If 'enable' is True then Enable Mode will be entered using 'enable_password'
    before any commands are executed.
    The telnet connection will attempt to stay open after the command is executed
    to allow things like multi-line commands to work properly.
    """
    import telnetlib
    global telnet_connection
    global prompt
    output = ""

    # Various daemon default ports:
    port_dict = {
        'zebra': 2601,
        'ripd': 2602,
        'ripngd': 2603,
        'ospfd': 2604,
        'bgpd': 2605,
        'ospf6d': 2606,
        'ospfapi': 2607,
        'isisd': 2608
    }

    if daemon: # Try to determine the port automatically
        try:
            port = port_dict[daemon]
        except Exception as e:
            return "Got exception trying to get a default port for '%s'.<br /> %s" % (daemon, e)

    if telnet_connection and not telnet_connection.eof: # A connection is open
        # Test to make sure the connection didn't timeout on the quagga side of things:
        telnet_connection.write("\n")
        telnet_connection.read_until(prompt, timeout)
    if not telnet_connection or telnet_connection.eof: # if there's no open connection, open one
        telnet_connection = telnetlib.Telnet(host, port, timeout)
        telnet_connection.read_until("Password:", timeout)
        telnet_connection.write(password + "\n")
        prompt = telnet_connection.read_until(">", timeout).strip()
        if 'Password:' in prompt:
            return "Bad password"
    if enable:
        if not prompt.endswith('#'):
            telnet_connection.write("enable\n")
            telnet_connection.read_until("Password:", timeout)
            telnet_connection.write(enable_password + "\n")
            prompt = telnet_connection.read_until("#", timeout).strip()
        if 'Password:' in prompt:
            return "Bad enable password"
    output += prompt
    telnet_connection.write(command + "\n")
    output += telnet_connection.read_until(prompt, timeout).rstrip(prompt)
    return output.lstrip() #.lstrip(command + '\n')